@@ -1,3 +1,7 @@
+#ifndef UNICODE
+#define UNICODE
+#endif
+
 #include <winsock2.h>
 #include <ws2bth.h>
 #include <BluetoothAPIs.h>
@@ -5,6 +9,23 @@
 #include <Python.h>
 
 #include <initguid.h>
+#include <port3.h>
+#include <wchar.h>
+
+#if 1
+static void dbg(const char *fmt, ...)
+{
+}
+#else
+static void dbg(const char *fmt, ...)
+{
+    va_list ap;
+    va_start (ap, fmt);
+    vprintf (fmt, ap);
+    va_end (ap);
+}
+#endif
+
 
 static void Err_SetFromWSALastError(PyObject *exc)
 {
@@ -12,7 +33,7 @@ static void Err_SetFromWSALastError(PyObject *exc)
 	FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM | FORMAT_MESSAGE_IGNORE_INSERTS,
 		NULL, WSAGetLastError(), 0, (LPTSTR) &lpMsgBuf, 0, NULL );
     PyErr_SetString( exc, lpMsgBuf );
-	free(lpMsgBuf);
+	LocalFree(lpMsgBuf);
 }
 
 
@@ -20,14 +41,14 @@ static void Err_SetFromWSALastError(PyObject *exc)
         if( !(cond) ) { Err_SetFromWSALastError( PyExc_IOError ); return 0; }
 
 static void
-ba2str( BTH_ADDR ba, char *addr )
+ba2str( BTH_ADDR ba, char *addr, size_t len )
 {
     int i;
     unsigned char bytes[6];
     for( i=0; i<6; i++ ) {
         bytes[5-i] = (unsigned char) ((ba >> (i*8)) & 0xff);
     }
-    sprintf(addr, "%02X:%02X:%02X:%02X:%02X:%02X", 
+    sprintf_s( addr, len, "%02X:%02X:%02X:%02X:%02X:%02X",
             bytes[0], bytes[1], bytes[2], bytes[3], bytes[4], bytes[5] );
 }
 
@@ -44,8 +65,7 @@ msbt_initwinsock(PyObject *self)
         return 0;
     }
 
-    Py_INCREF( Py_None );
-    return Py_None;
+    Py_RETURN_NONE;
 }
 PyDoc_STRVAR(msbt_initwinsock_doc, "TODO");
 
@@ -87,20 +107,20 @@ str2uuid( const char *uuid_str, GUID *uuid)
     int i;
     char buf[20] = { 0 };
 
-    strncpy(buf, uuid_str, 8);
+    strncpy_s(buf, _countof(buf), uuid_str, 8);
     uuid->Data1 = strtoul( buf, NULL, 16 );
     memset(buf, 0, sizeof(buf));
 
-    strncpy(buf, uuid_str+9, 4);
+    strncpy_s(buf, _countof(buf), uuid_str+9, 4);
     uuid->Data2 = (unsigned short) strtoul( buf, NULL, 16 );
     memset(buf, 0, sizeof(buf));
     
-    strncpy(buf, uuid_str+14, 4);
+    strncpy_s(buf, _countof(buf), uuid_str+14, 4);
     uuid->Data3 = (unsigned short) strtoul( buf, NULL, 16 );
     memset(buf, 0, sizeof(buf));
 
-    strncpy(buf, uuid_str+19, 4);
-    strncpy(buf+4, uuid_str+24, 12);
+    strncpy_s(buf, _countof(buf), uuid_str+19, 4);
+    strncpy_s(buf+4, _countof(buf)-4, uuid_str+24, 12);
     for( i=0; i<8; i++ ) {
         char buf2[3] = { buf[2*i], buf[2*i+1], 0 };
         uuid->Data4[i] = (unsigned char)strtoul( buf2, NULL, 16 );
@@ -134,7 +154,7 @@ PyDoc_STRVAR(msbt_socket_doc, "TODO");
 static PyObject *
 msbt_bind(PyObject *self, PyObject *args)
 {
-    char *addrstr = NULL;
+    wchar_t *addrstr = NULL;
     int addrstrlen = -1;
     int sockfd = -1;
     int port = -1;
@@ -163,9 +183,8 @@ msbt_bind(PyObject *self, PyObject *args)
 
     _CHECK_OR_RAISE_WSA( NO_ERROR == status );
 
-    Py_INCREF(Py_None);
-    return Py_None;
-};
+    Py_RETURN_NONE;
+}
 PyDoc_STRVAR(msbt_bind_doc, "TODO");
 
 static PyObject *
@@ -182,9 +201,8 @@ msbt_listen(PyObject *self, PyObject *args)
 
     _CHECK_OR_RAISE_WSA( NO_ERROR == status );
     
-    Py_INCREF(Py_None);
-    return Py_None;
-};
+    Py_RETURN_NONE;
+}
 PyDoc_STRVAR(msbt_listen_doc, "TODO");
 
 static PyObject *
@@ -206,7 +224,7 @@ msbt_accept(PyObject *self, PyObject *args)
 
     _CHECK_OR_RAISE_WSA( SOCKET_ERROR != clientfd );
 
-    ba2str(ca.btAddr, buf);
+    ba2str(ca.btAddr, buf, _countof(buf));
     result = Py_BuildValue( "isi", clientfd, buf, ca.port );
     return result;
 };
@@ -216,7 +234,7 @@ static PyObject *
 msbt_connect(PyObject *self, PyObject *args)
 {
     int sockfd = -1;
-    char *addrstr = NULL;
+    wchar_t *addrstr = NULL;
     int port = -1;
     SOCKADDR_BTH sa = { 0 };
     int sa_len = sizeof(sa);
@@ -238,9 +256,8 @@ msbt_connect(PyObject *self, PyObject *args)
 
     _CHECK_OR_RAISE_WSA( NO_ERROR == status );
 
-    Py_INCREF(Py_None);
-    return Py_None;
-};
+    Py_RETURN_NONE;
+}
 PyDoc_STRVAR(msbt_connect_doc, "TODO");
 
 static PyObject *
@@ -259,12 +276,6 @@ msbt_send(PyObject *self, PyObject *args)
     sent = send(sockfd, data, datalen, flags);
     Py_END_ALLOW_THREADS;
 
-    if (WSAGetLastError() == WSAEWOULDBLOCK ||
-        WSAGetLastError() == WSAETIMEDOUT ) {
-		return PyInt_FromLong ( 0 );
-	}
-
-
     _CHECK_OR_RAISE_WSA( SOCKET_ERROR != sent );
     
     return PyInt_FromLong( sent );
@@ -288,17 +299,10 @@ msbt_recv(PyObject *self, PyObject *args)
     received = recv(sockfd, PyString_AS_STRING(buf), datalen, flags);
     Py_END_ALLOW_THREADS;
 
-    if (WSAGetLastError() == WSAEWOULDBLOCK ||
-        WSAGetLastError() == WSAETIMEDOUT ) {
-        _PyString_Resize(&buf, 0); // XXX is this necessary? -albert
-
-        return buf;
-    }
-
     if( SOCKET_ERROR == received ){
         Py_DECREF(buf);
-        return 0;   
     }
+    _CHECK_OR_RAISE_WSA( SOCKET_ERROR != received );
     if( received != datalen ) _PyString_Resize(&buf, received);
 
     return buf;
@@ -316,18 +320,49 @@ msbt_close(PyObject *self, PyObject *args)
     status = closesocket(sockfd);
     Py_END_ALLOW_THREADS;
 
-    Py_INCREF(Py_None);
-    return Py_None;
-};
+    Py_RETURN_NONE;
+}
 PyDoc_STRVAR(msbt_close_doc, "TODO");
 
+static PyObject *
+msbt_getpeername(PyObject *self, PyObject *args)
+{
+    int sockfd = -1;
+    SOCKADDR_BTH sa = { 0 };
+    int sa_len = sizeof(sa);
+    char buf[18] = { 0 };
+    int buf_len = sizeof(buf);
+    int status;
+    if(!PyArg_ParseTuple( args, "i", &sockfd )) return 0;
+
+    sa.addressFamily = AF_BTH;
+    Py_BEGIN_ALLOW_THREADS;
+    status = getpeername( sockfd, (LPSOCKADDR)&sa, &sa_len );
+    Py_END_ALLOW_THREADS;
+
+    _CHECK_OR_RAISE_WSA( NO_ERROR == status );
+
+    ba2str( sa.btAddr, buf, buf_len );
+    return Py_BuildValue( "si", buf, sa.port );
+};
+PyDoc_STRVAR(msbt_getpeername_doc,
+"getpeername() -> address info\n\
+\n\
+Return the address of the peer to which a socket is connected.\n\
+This function works with any address family and it simply returns the \n\
+address to which the socket is connected. The getpeername function can be \n\
+used only on a connected socket.\n\
+For datagram sockets, only the address of a peer specified in a previous \n\
+connect call will be returned. Any address specified by a previous sendto \n\
+call will not be returned by getpeername.");
+
 static PyObject *
 msbt_getsockname(PyObject *self, PyObject *args)
 {
     int sockfd = -1;
     SOCKADDR_BTH sa = { 0 };
     int sa_len = sizeof(sa);
-    char buf[100] = { 0 };
+    char buf[18] = { 0 };
     int buf_len = sizeof(buf);
     int status;
     if(!PyArg_ParseTuple( args, "i", &sockfd )) return 0;
@@ -339,7 +374,7 @@ msbt_getsockname(PyObject *self, PyObject *args)
 
     _CHECK_OR_RAISE_WSA( NO_ERROR == status );
 
-    ba2str( sa.btAddr, buf );
+    ba2str( sa.btAddr, buf, buf_len );
     return Py_BuildValue( "si", buf, sa.port );
 };
 PyDoc_STRVAR(msbt_getsockname_doc, "TODO");
@@ -372,94 +407,138 @@ msbt_dup(PyObject *self, PyObject *args)
 PyDoc_STRVAR(msbt_dup_doc, "TODO");
 
 // ====================
+static int bt_adapter_present()
+{
+   int hwactive = 1;
+   SOCKADDR_BTH sa = { 0 };
+
+   int s = socket(AF_BTH, SOCK_STREAM, BTHPROTO_RFCOMM);
+
+   sa.addressFamily = AF_BTH;
+   sa.port = BT_PORT_ANY;
+   if (s == INVALID_SOCKET)
+   {
+      closesocket(s);
+      return 0;
+   }
+
+   if (SOCKET_ERROR == bind(s, (LPSOCKADDR)&sa, sizeof(sa)))
+   {
+	   hwactive = 0;
+   }
+
+   closesocket(s);
+   return hwactive;
+}
 
 static PyObject *
-msbt_discover_devices(PyObject *self, PyObject *args)
+msbt_discover_devices(PyObject *self, PyObject *args, PyObject *kwds)
 {
-    int flushcache = 0;
-    int lookupnames = 0;
+    BLUETOOTH_DEVICE_INFO device_info;
+    BLUETOOTH_DEVICE_SEARCH_PARAMS search_criteria;
+    HBLUETOOTH_DEVICE_FIND found_device;
+    BOOL next = TRUE;
+    PyObject * toreturn = NULL;
+    int duration = 8;
+    int flush_cache = 1;
 
-	// inquiry data structure
-	DWORD qs_len = sizeof( WSAQUERYSET );
-	WSAQUERYSET *qs = (WSAQUERYSET*) malloc( qs_len );
+    static char *keywords[] = {"duration", "flush_cache", 0};
+    if(!PyArg_ParseTupleAndKeywords(args, kwds, "ii", keywords,
+                &duration, &flush_cache)) {
+        return NULL;
+    }
+    toreturn = PyList_New(0);
+    ZeroMemory(&device_info, sizeof(BLUETOOTH_DEVICE_INFO));
+    ZeroMemory(&search_criteria, sizeof(BLUETOOTH_DEVICE_SEARCH_PARAMS));
 
-	DWORD flags = LUP_CONTAINERS;
-	HANDLE h;
-	int done = 0;
-    PyObject *tup = NULL;
-    PyObject *toreturn = NULL;
+    device_info.dwSize = sizeof(BLUETOOTH_DEVICE_INFO);
+    search_criteria.dwSize = sizeof(BLUETOOTH_DEVICE_SEARCH_PARAMS);
+    search_criteria.fReturnAuthenticated = TRUE;
+    search_criteria.fReturnRemembered = !flush_cache;
+    search_criteria.fReturnConnected = TRUE;
+    search_criteria.fReturnUnknown = TRUE;
+    search_criteria.fIssueInquiry = TRUE;
+    search_criteria.cTimeoutMultiplier = duration;
+    search_criteria.hRadio = NULL;
 
-    if(!PyArg_ParseTuple(args, "ii", &flushcache, &lookupnames)) {
-        free( qs );
-        return 0;
+    Py_BEGIN_ALLOW_THREADS;
+    found_device = BluetoothFindFirstDevice(&search_criteria, &device_info);
+    Py_END_ALLOW_THREADS;
+
+    _CHECK_OR_RAISE_WSA(found_device != NULL)
+
+    while(next) {
+        PyObject *tup = NULL;
+        PyObject *item_tuple = NULL;
+        char buf[40] = {0};
+        BTH_ADDR result = device_info.Address.ullLong;
+        ba2str( result, buf, _countof(buf) );
+
+        tup = PyTuple_New(3);
+        item_tuple = PyString_FromString(buf);
+        PyTuple_SetItem( tup, 0, item_tuple );
+        item_tuple = PyUnicode_FromUnicode(
+                (const Py_UNICODE*)device_info.szName,
+                wcslen(device_info.szName));
+        PyTuple_SetItem( tup, 1, item_tuple );
+        item_tuple = PyInt_FromLong(device_info.ulClassofDevice);
+        PyTuple_SetItem( tup, 2, item_tuple );
+        PyList_Append( toreturn, tup );
+        Py_DECREF( tup );
+        
+        Py_BEGIN_ALLOW_THREADS;
+        next = BluetoothFindNextDevice(found_device, &device_info);
+        Py_END_ALLOW_THREADS;
+    }
+    return toreturn;
+}
+PyDoc_STRVAR(msbt_discover_devices_doc,
+	"msbt_discover_devices(duration=8, flush_cache=True\n\\n\
+	Performs a device inquiry. The inquiry will last 1.28 * duration seconds.\
+	If flush_cache is True, then new inquiry will be performed,\
+	else cashed devices will be returned(plus paired devices).)");
+
+
+static PyObject *
+msbt_list_local(PyObject *self)
+{
+    HANDLE m_radio = NULL;
+    HBLUETOOTH_RADIO_FIND m_bt = NULL;
+    BOOL next = TRUE;
+    BLUETOOTH_FIND_RADIO_PARAMS
+            find_radio = {sizeof(BLUETOOTH_FIND_RADIO_PARAMS)};
+    PyObject * toreturn = PyList_New(0);
+
+    Py_BEGIN_ALLOW_THREADS;
+    m_bt = BluetoothFindFirstRadio(&find_radio, &m_radio);
+    Py_END_ALLOW_THREADS;
+    _CHECK_OR_RAISE_WSA(m_bt != NULL);
+
+    while(next) {
+        PyObject *item = NULL;
+        char buf[40] = {0};
+        BLUETOOTH_RADIO_INFO m_bt_info = {sizeof(BLUETOOTH_RADIO_INFO), 0,};
+
+        // Then get the radio device info....
+        DWORD mbtinfo_ret = BluetoothGetRadioInfo(m_radio, &m_bt_info);
+        _CHECK_OR_RAISE_WSA(mbtinfo_ret == ERROR_SUCCESS);
+
+        ba2str(m_bt_info.address.ullLong, buf, _countof(buf));
+        item = PyString_FromString(buf);
+        PyList_Append(toreturn, item);
+        Py_DECREF(item);
+
+        next = BluetoothFindNextRadio(&find_radio, &m_radio);
     }
 
-    if (flushcache) flushcache = LUP_FLUSHCACHE;
-    if (lookupnames) lookupnames = LUP_RETURN_NAME;
-
-	ZeroMemory( qs, qs_len );
-	qs->dwSize = sizeof(WSAQUERYSET);
-	qs->dwNameSpace = NS_BTH;
-
-	flags |= flushcache | lookupnames | LUP_RETURN_ADDR;
-
-	// start the device inquiry
-	if( SOCKET_ERROR == WSALookupServiceBegin( qs, flags, &h )) {
-        int error = WSAGetLastError();
-        if( error == WSASERVICE_NOT_FOUND ) {
-            // no devices detected.  return an empty list
-            WSALookupServiceEnd( h );
-            free( qs );
-            return PyList_New(0);
-        } else {
-            Err_SetFromWSALastError( PyExc_IOError );
-            free(qs);
-            return 0;
-        }
-	}
-
-    toreturn = PyList_New(0);
-
-	// iterate through the inquiry results
-	while(! done) {
-		if( NO_ERROR == WSALookupServiceNext( h, flags, &qs_len, qs )) {
-			char buf[40] = {0};
-			BTH_ADDR result = 
-				((SOCKADDR_BTH*)qs->lpcsaBuffer->RemoteAddr.lpSockaddr)->btAddr;
-            ba2str( result, buf );
-
-            if( lookupnames ) {
-                tup = Py_BuildValue( "ss", buf, qs->lpszServiceInstanceName );
-            } else {
-                tup = Py_BuildValue( "s", buf );
-            }
-            PyList_Append( toreturn, tup );
-            Py_DECREF( tup );
-		} else {
-			int error = WSAGetLastError();
-			
-			if( error == WSAEFAULT ) {
-                // the qs data structure is too small.  allocate a bigger
-                // buffer and try again.
-				free( qs );
-				qs = (WSAQUERYSET*) malloc( qs_len );
-			} else if( error == WSA_E_NO_MORE ) {
-                // no more results.
-				done = 1;
-			} else {
-                // unexpected error.  raise an exception.
-                Err_SetFromWSALastError( PyExc_IOError );
-                Py_DECREF( toreturn );
-                return 0;
-			}
-		}
-	}
-	WSALookupServiceEnd( h );
-	free( qs );
+    // No more radio, close the radio handle
+    _CHECK_OR_RAISE_WSA(BluetoothFindRadioClose(m_bt) == TRUE);
 
     return toreturn;
 }
-PyDoc_STRVAR(msbt_discover_devices_doc, "TODO");
+PyDoc_STRVAR(msbt_list_local_doc,
+    "msbt_list_local - List local BT adapters addresses");
+
 
 static PyObject *
 msbt_lookup_name(PyObject *self, PyObject *args)
@@ -468,7 +547,7 @@ msbt_lookup_name(PyObject *self, PyObject *args)
     BLUETOOTH_FIND_RADIO_PARAMS p = { sizeof(p) };
     HBLUETOOTH_RADIO_FIND fhandle = NULL;
     BLUETOOTH_DEVICE_INFO dinfo = { 0 };
-    char *addrstr = NULL;
+    wchar_t *addrstr = NULL;
     SOCKADDR_BTH sa = { 0 };
     int sa_len = sizeof(sa);
     DWORD status;
@@ -511,6 +590,7 @@ msbt_find_service(PyObject *self, PyObject *args)
 	DWORD flags = LUP_FLUSHCACHE | LUP_RETURN_ALL;
 	HANDLE h;
 	int done = 0;
+    int status = 0;
     PyObject *record = NULL;
     PyObject *toreturn = NULL;
     GUID uuid = { 0 };
@@ -523,7 +603,7 @@ msbt_find_service(PyObject *self, PyObject *args)
 	qs->dwSize = sizeof(WSAQUERYSET);
 	qs->dwNameSpace = NS_BTH;
     qs->dwNumberOfCsAddrs = 0;
-    qs->lpszContext = (LPSTR) localAddressBuf;
+    qs->lpszContext = (LPWSTR) localAddressBuf;
 
     if( 0 == strcmp( addrstr, "localhost" ) ) {
         // find the Bluetooth address of the first local adapter. 
@@ -547,7 +627,7 @@ msbt_find_service(PyObject *self, PyObject *args)
         _CHECK_OR_RAISE_WSA( \
             ERROR_SUCCESS != BluetoothGetRadioInfo( rhandle, &info ) );
 
-        ba2str( info.address.ullLong, localAddressBuf );
+        ba2str( info.address.ullLong, localAddressBuf, _countof(localAddressBuf) );
 #else
         // bind a temporary socket and get its Bluetooth address
         SOCKADDR_BTH sa = { 0 };
@@ -562,13 +642,13 @@ msbt_find_service(PyObject *self, PyObject *args)
         _CHECK_OR_RAISE_WSA(NO_ERROR == getsockname(tmpfd, (LPSOCKADDR)&sa,\
                     &sa_len ) );
 
-        ba2str(sa.btAddr, localAddressBuf );
-        close(tmpfd);
+        ba2str(sa.btAddr, localAddressBuf, _countof(localAddressBuf) );
+        _close(tmpfd);
 #endif
 
         flags |= LUP_RES_SERVICE;
     } else {
-        strcpy( localAddressBuf, addrstr );
+        strcpy_s(localAddressBuf, _countof(localAddressBuf), addrstr);
     }
 
     if( strlen(uuidstr) != 36 || uuidstr[8] != '-' || uuidstr[13] != '-' 
@@ -580,7 +660,10 @@ msbt_find_service(PyObject *self, PyObject *args)
     str2uuid( uuidstr, &uuid );
     qs->lpServiceClassId = &uuid;
 
-	if( SOCKET_ERROR == WSALookupServiceBegin( qs, flags, &h )) {
+    Py_BEGIN_ALLOW_THREADS;
+    status = WSALookupServiceBegin( qs, flags, &h );
+    Py_END_ALLOW_THREADS;
+	if( SOCKET_ERROR == status) {
         int err_code = WSAGetLastError();
         if( WSASERVICE_NOT_FOUND == err_code ) {
             // this device does not advertise any services.  return an
@@ -599,7 +682,10 @@ msbt_find_service(PyObject *self, PyObject *args)
 
 	// iterate through the inquiry results
 	while(! done) {
-		if( NO_ERROR == WSALookupServiceNext( h, flags, &qs_len, qs )) {
+        Py_BEGIN_ALLOW_THREADS;
+        status = WSALookupServiceNext( h, flags, &qs_len, qs );
+        Py_END_ALLOW_THREADS;
+		if (NO_ERROR == status) {
             int proto;
             int port;
             PyObject *rawrecord = NULL;
@@ -610,10 +696,10 @@ msbt_find_service(PyObject *self, PyObject *args)
             dict_set_strings( record, "host", localAddressBuf );
             
             // set service name
-            dict_set_strings( record, "name", qs->lpszServiceInstanceName );
+            dict_set_strings( record, "name", (const char*) qs->lpszServiceInstanceName );
 
             // set description
-            dict_set_strings( record, "description", qs->lpszComment );
+            dict_set_strings( record, "description", (const char*) qs->lpszComment );
 
             // set protocol and port
             csinfo = qs->lpcsaBuffer;
@@ -662,7 +748,9 @@ msbt_find_service(PyObject *self, PyObject *args)
 			}
 		}
 	}
+    Py_BEGIN_ALLOW_THREADS;
 	WSALookupServiceEnd( h );
+    Py_END_ALLOW_THREADS;
 	free( qs );
 
     return toreturn;
@@ -720,6 +808,8 @@ msbt_set_service_raw(PyObject *self, PyObject *args)
 }
 PyDoc_STRVAR(msbt_set_service_raw_doc, "");
 
+
+
 static PyObject *
 msbt_set_service(PyObject *self, PyObject *args)
 {
@@ -729,8 +819,8 @@ msbt_set_service(PyObject *self, PyObject *args)
 
 	SOCKADDR_BTH sa = { 0 };
 	int sa_len = sizeof(sa);
-    char *service_name = NULL;
-    char *service_desc = NULL;
+    wchar_t *service_name = NULL;
+    wchar_t *service_desc = NULL;
     char *service_class_id_str = NULL;
 	CSADDR_INFO sockInfo = { 0 };
     GUID uuid = { 0 };
@@ -767,9 +857,7 @@ msbt_set_service(PyObject *self, PyObject *args)
         Err_SetFromWSALastError( PyExc_IOError );
         return 0;
     }
-
-    Py_INCREF( Py_None );
-    return Py_None;
+    Py_RETURN_NONE;
 }
 PyDoc_STRVAR(msbt_set_service_doc, "");
 
@@ -778,25 +866,12 @@ msbt_setblocking(PyObject *self, PyObject *args)
 {
     int sockfd = -1;
     int block = -1;
-    PyObject* blockingArg;
 
-    if(!PyArg_ParseTuple(args, "iO", &sockfd, &blockingArg)) return 0;
+    if(!PyArg_ParseTuple(args, "ii", &sockfd, &block)) return 0;
+    block = !block; //set zero to non-zero and non-zero to zero
+    ioctlsocket( sockfd, FIONBIO, &block);
 
-    block = PyInt_AsLong(blockingArg);
-
-    if (block == -1 && PyErr_Occurred())
-        return NULL;
-
-    if (block) {
-        block = 0;
-        ioctlsocket( sockfd, FIONBIO, &block);
-    } else {
-        block = 1;
-        ioctlsocket( sockfd, FIONBIO, &block);
-    }
-
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 PyDoc_STRVAR(msbt_setblocking_doc, "");
 
@@ -824,8 +899,7 @@ msbt_settimeout(PyObject *self, PyObject *args)
         return 0;
     }
 
-    Py_INCREF(Py_None);
-    return Py_None;
+    Py_RETURN_NONE;
 }
 PyDoc_STRVAR(msbt_settimeout_doc, "");
 
@@ -851,6 +925,71 @@ msbt_gettimeout(PyObject *self, PyObject *args)
 }
 PyDoc_STRVAR(msbt_gettimeout_doc, "");
 
+/* s.setsockopt() method.
+   With an integer third argument, sets an integer option.
+   With a string third argument, sets an option from a buffer;
+   use optional built-in module 'struct' to encode the string. */
+
+static PyObject *
+msbt_setsockopt(PyObject *s, PyObject *args)
+{
+    int sockfd = -1;
+    int level;
+    ULONG optname;
+    int res;
+    ULONG flag;
+
+    if (!PyArg_ParseTuple(args, "iiIi:setsockopt", &sockfd, &level, &optname,
+            &flag)) {
+        return 0;
+    }
+    res = setsockopt(sockfd, level, optname, (char*) &flag, sizeof(ULONG));
+    if (res < 0) {
+        Err_SetFromWSALastError(PyExc_IOError);
+        return 0;
+    }
+    Py_RETURN_NONE;
+}
+
+PyDoc_STRVAR(msbt_setsockopt_doc,
+"setsockopt(level, option, value)\n\
+\n\
+Set a socket option.  See the Unix manual for level and option.\n\
+The value argument can either be an integer or a string.");
+
+
+/* s.getsockopt() method.
+   With two arguments, retrieves an integer option.
+   With a third integer argument, retrieves a string buffer of that size;
+   use optional built-in module 'struct' to decode the string. */
+
+static PyObject *
+msbt_getsockopt(PyObject *s, PyObject *args)
+{
+    int sockfd = -1;
+    int level;
+    ULONG optname;
+    int res;
+    ULONG flag;
+    int flagsize = sizeof flag;
+
+    if (!PyArg_ParseTuple(args, "iiI", &sockfd, &level, &optname))
+        return 0;
+
+    res = getsockopt(sockfd, level, optname, (void *) &flag, &flagsize);
+    if (res < 0) {
+        Err_SetFromWSALastError(PyExc_IOError);
+        return 0;
+    }
+    return PyInt_FromLong(flag);
+}
+
+PyDoc_STRVAR(msbt_getsockopt_doc,
+"getsockopt(level, option[, buffersize]) -> value\n\
+\n\
+Get a socket option.  See the Unix manual for level and option.\n\
+If a nonzero buffersize argument is given, the return value is a\n\
+string of that length; otherwise it is an integer.");
 
 
 // =======================  ADMINISTRATIVE =========================
@@ -868,8 +1007,13 @@ static PyMethodDef msbt_methods[] = {
     { "close", (PyCFunction)msbt_close, METH_VARARGS, msbt_close_doc },
     { "getsockname", (PyCFunction)msbt_getsockname, METH_VARARGS, 
         msbt_getsockname_doc },
+    { "getpeername", (PyCFunction)msbt_getpeername, METH_VARARGS,
+        msbt_getpeername_doc },
     { "dup", (PyCFunction)msbt_dup, METH_VARARGS, msbt_dup_doc },
-    { "discover_devices", (PyCFunction)msbt_discover_devices, METH_VARARGS, msbt_discover_devices_doc },
+    { "discover_devices", (PyCFunction)msbt_discover_devices,
+            METH_VARARGS | METH_KEYWORDS, msbt_discover_devices_doc },
+    { "list_local", (PyCFunction)msbt_list_local,
+            METH_NOARGS, msbt_list_local_doc },
     { "lookup_name", (PyCFunction)msbt_lookup_name, METH_VARARGS, msbt_lookup_name_doc },
     { "find_service", (PyCFunction)msbt_find_service, METH_VARARGS, msbt_find_service_doc },
     { "set_service", (PyCFunction)msbt_set_service, METH_VARARGS, msbt_set_service_doc },
@@ -877,21 +1021,53 @@ static PyMethodDef msbt_methods[] = {
     { "setblocking", (PyCFunction)msbt_setblocking, METH_VARARGS, msbt_setblocking_doc },
     { "settimeout", (PyCFunction)msbt_settimeout, METH_VARARGS, msbt_settimeout_doc },
     { "gettimeout", (PyCFunction)msbt_gettimeout, METH_VARARGS, msbt_gettimeout_doc },
+    { "setsockopt", (PyCFunction)msbt_setsockopt, METH_VARARGS, msbt_setsockopt_doc },
+    { "getsockopt", (PyCFunction)msbt_getsockopt, METH_VARARGS, msbt_getsockopt_doc },
     { NULL, NULL }
 };
 
 PyDoc_STRVAR(msbt_doc, "TODO\n");
 
+#define ADD_INT_CONSTANT(m,a) PyModule_AddIntConstant(m, #a, a)
+
+#if PY_MAJOR_VERSION < 3
 PyMODINIT_FUNC
 init_msbt(void)
+{
+    PyObject * m = Py_InitModule3("_msbt", msbt_methods, msbt_doc);
+#else
+PyMODINIT_FUNC
+PyInit__msbt(void)
 {
     PyObject *m;
 
-    m = Py_InitModule3("_msbt", msbt_methods, msbt_doc);
-
-#define ADD_INT_CONSTANT(m,a) PyModule_AddIntConstant(m, #a, a)
+    static struct PyModuleDef moduledef = {
+        PyModuleDef_HEAD_INIT,
+        "_msbt",
+        NULL,
+        -1,
+        msbt_methods,
+        NULL,
+        NULL,
+        NULL,
+        NULL
+    };
+    m = PyModule_Create(&moduledef);
+#endif
 
+    ADD_INT_CONSTANT(m, AF_BTH);
     ADD_INT_CONSTANT(m, SOCK_STREAM);
     ADD_INT_CONSTANT(m, BTHPROTO_RFCOMM);
     ADD_INT_CONSTANT(m, BT_PORT_ANY);
+
+    ADD_INT_CONSTANT(m, SOL_RFCOMM);
+    ADD_INT_CONSTANT(m, SO_BTH_AUTHENTICATE);
+    ADD_INT_CONSTANT(m, SO_BTH_ENCRYPT);
+    ADD_INT_CONSTANT(m, SO_BTH_MTU);
+    ADD_INT_CONSTANT(m, SO_BTH_MTU_MAX);
+    ADD_INT_CONSTANT(m, SO_BTH_MTU_MIN);
+
+#if PY_MAJOR_VERSION >= 3
+    return m;
+#endif
 }
